@@ -3,26 +3,31 @@
   dln.c -
 
   $Author: nobu $
-  $Date: 2008-01-11 11:50:38 +0900 (Fri, 11 Jan 2008) $
   created at: Tue Jan 18 17:05:06 JST 1994
 
-  Copyright (C) 1993-2003 Yukihiro Matsumoto
+  Copyright (C) 1993-2007 Yukihiro Matsumoto
 
 **********************************************************************/
 
-#include "ruby.h"
+#ifdef RUBY_EXPORT
+#include "ruby/ruby.h"
+#define dln_notimplement rb_notimplement
+#define dln_memerror rb_memerror
+#define dln_exit rb_exit
+#define dln_loaderror rb_loaderror
+#else
+#define dln_notimplement --->>> dln not implemented <<<---
+#define dln_memerror abort
+#define dln_exit exit
+static void dln_loaderror(const char *format, ...);
+#endif
 #include "dln.h"
+#include "internal.h"
 
 #ifdef HAVE_STDLIB_H
 # include <stdlib.h>
 #endif
 
-#ifdef __CHECKER__
-#undef HAVE_DLOPEN
-#undef USE_DLN_A_OUT
-#undef USE_DLN_DLOPEN
-#endif
-
 #ifdef USE_DLN_A_OUT
 char *dln_argv0;
 #endif
@@ -43,15 +48,18 @@ void *xcalloc();
 void *xrealloc();
 #endif
 
+#undef free
+#define free(x) xfree(x)
+
 #include <stdio.h>
-#if defined(_WIN32) || defined(__VMS)
+#if defined(_WIN32)
 #include "missing/file.h"
 #endif
 #include <sys/types.h>
 #include <sys/stat.h>
 
 #ifndef S_ISDIR
-#   define S_ISDIR(m) ((m & S_IFMT) == S_IFDIR)
+#   define S_ISDIR(m) (((m) & S_IFMT) == S_IFDIR)
 #endif
 
 #ifdef HAVE_SYS_PARAM_H
@@ -69,19 +77,7 @@ void *xrealloc();
 char *getenv();
 #endif
 
-#if defined(__VMS)
-#pragma builtins
-#include <dlfcn.h>
-#endif
-
-#ifdef __MACOS__
-# include <TextUtils.h>
-# include <CodeFragments.h>
-# include <Aliases.h>
-# include "macruby_private.h"
-#endif
-
-#if defined(__APPLE__) && defined(__MACH__)   /* Mac OS X */
+#ifdef __APPLE__
 # if defined(HAVE_DLOPEN)
    /* Mac OS X with dlopen (10.3 or later) */
 #  define MACOSX_DLOPEN
@@ -90,63 +86,65 @@ char *getenv();
 # endif
 #endif
 
-#ifdef __BEOS__
-# include <image.h>
+#ifndef dln_loaderror
+static void
+dln_loaderror(const char *format, ...)
+{
+    va_list ap;
+    va_start(ap, format);
+    vfprintf(stderr, format, ap);
+    va_end(ap);
+    abort();
+}
 #endif
 
-#ifndef NO_DLN_LOAD
-
 #if defined(HAVE_DLOPEN) && !defined(USE_DLN_A_OUT) && !defined(_AIX) && !defined(MACOSX_DYLD) && !defined(_UNICOSMP)
 /* dynamic load with dlopen() */
 # define USE_DLN_DLOPEN
 #endif
 
-#ifndef FUNCNAME_PATTERN
-# if defined(__hp9000s300) ||  (defined(__NetBSD__) && !defined(__ELF__)) || defined(__BORLANDC__) || (defined(__FreeBSD__) && !defined(__ELF__)) || (defined(__OpenBSD__) && !defined(__ELF__)) || defined(NeXT) || defined(__WATCOMC__) || defined(MACOSX_DYLD)
-#  define FUNCNAME_PATTERN "_Init_%s"
-# else
-#  define FUNCNAME_PATTERN "Init_%s"
-# endif
+#if defined(__hp9000s300) || ((defined(__NetBSD__) || defined(__FreeBSD__) || defined(__OpenBSD__)) && !defined(__ELF__)) || defined(NeXT) || defined(MACOSX_DYLD)
+# define EXTERNAL_PREFIX "_"
+#else
+# define EXTERNAL_PREFIX ""
+#endif
+#define FUNCNAME_PREFIX EXTERNAL_PREFIX"Init_"
+
+#if defined __CYGWIN__ || defined DOSISH
+#define isdirsep(x) ((x) == '/' || (x) == '\\')
+#else
+#define isdirsep(x) ((x) == '/')
 #endif
 
-static int
-init_funcname_len(buf, file)
-    char **buf;
-    const char *file;
+static size_t
+init_funcname_len(const char **file)
 {
-    char *p;
-    const char *slash;
-    int len;
+    const char *p = *file, *base, *dot = NULL;
 
     /* Load the file as an object one */
-    for (slash = file-1; *file; file++) /* Find position of last '/' */
-#ifdef __MACOS__
-	if (*file == ':') slash = file;
-#else
-	if (*file == '/') slash = file;
-#endif
-
-    len = strlen(FUNCNAME_PATTERN) + strlen(slash + 1);
-    *buf = xmalloc(len);
-    snprintf(*buf, len, FUNCNAME_PATTERN, slash + 1);
-    for (p = *buf; *p; p++) {         /* Delete suffix if it exists */
-	if (*p == '.') {
-	    *p = '\0'; break;
-	}
+    for (base = p; *p; p++) { /* Find position of last '/' */
+	if (*p == '.' && !dot) dot = p;
+	if (isdirsep(*p)) base = p+1, dot = NULL;
     }
-    return p - *buf;
+    *file = base;
+    /* Delete suffix if it exists */
+    return (dot ? dot : p) - base;
 }
 
+static const char funcname_prefix[sizeof(FUNCNAME_PREFIX) - 1] = FUNCNAME_PREFIX;
+
 #define init_funcname(buf, file) do {\
-    int len = init_funcname_len(buf, file);\
-    char *tmp = ALLOCA_N(char, len+1);\
+    const char *base = (file);\
+    const size_t flen = init_funcname_len(&base);\
+    const size_t plen = sizeof(funcname_prefix);\
+    char *const tmp = ALLOCA_N(char, plen+flen+1);\
     if (!tmp) {\
-	free(*buf);\
-	rb_memerror();\
+	dln_memerror();\
     }\
-    strcpy(tmp, *buf);\
-    free(*buf);\
-    *buf = tmp;\
+    memcpy(tmp, funcname_prefix, plen);\
+    memcpy(tmp+plen, base, flen);\
+    tmp[plen+flen] = '\0';\
+    *(buf) = tmp;\
 } while (0)
 
 #ifdef USE_DLN_A_OUT
@@ -184,8 +182,8 @@ static int dln_init_p = 0;
 
 #define INVALID_OBJECT(h) (N_MAGIC(h) != OMAGIC)
 
-#include "util.h"
-#include "st.h"
+#include "ruby/util.h"
+#include "ruby/st.h"
 
 static st_table *sym_tbl;
 static st_table *undef_tbl;
@@ -193,10 +191,7 @@ static st_table *undef_tbl;
 static int load_lib();
 
 static int
-load_header(fd, hdrp, disp)
-    int fd;
-    struct exec *hdrp;
-    long disp;
+load_header(int fd, struct exec *hdrp, long disp)
 {
     int size;
 
@@ -230,20 +225,20 @@ load_header(fd, hdrp, disp)
 #define RELOC_TARGET_SIZE(r)		((r)->r_length)
 #endif
 
-#if defined(sun) && defined(sparc)
+#if defined(__sun) && defined(__sparc)
 /* Sparc (Sun 4) macros */
 #  undef relocation_info
 #  define relocation_info reloc_info_sparc
 #  define R_RIGHTSHIFT(r)	(reloc_r_rightshift[(r)->r_type])
 #  define R_BITSIZE(r) 		(reloc_r_bitsize[(r)->r_type])
 #  define R_LENGTH(r)		(reloc_r_length[(r)->r_type])
-static int reloc_r_rightshift[] = {
+static const int reloc_r_rightshift[] = {
   0, 0, 0, 0, 0, 0, 2, 2, 10, 0, 0, 0, 0, 0, 0,
 };
-static int reloc_r_bitsize[] = {
+static const int reloc_r_bitsize[] = {
   8, 16, 32, 8, 16, 32, 30, 22, 22, 22, 13, 10, 32, 32, 16,
 };
-static int reloc_r_length[] = {
+static const int reloc_r_length[] = {
   0, 1, 2, 0, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 };
 #  define R_PCREL(r) \
@@ -266,10 +261,7 @@ static int reloc_r_length[] = {
 #endif
 
 static struct relocation_info *
-load_reloc(fd, hdrp, disp)
-     int fd;
-     struct exec *hdrp;
-     long disp;
+load_reloc(int fd, struct exec *hdrp, long disp)
 {
     struct relocation_info *reloc;
     int size;
@@ -292,10 +284,7 @@ load_reloc(fd, hdrp, disp)
 }
 
 static struct nlist *
-load_sym(fd, hdrp, disp)
-    int fd;
-    struct exec *hdrp;
-    long disp;
+load_sym(int fd, struct exec *hdrp, long disp)
 {
     struct nlist * buffer;
     struct nlist * sym;
@@ -336,9 +325,7 @@ load_sym(fd, hdrp, disp)
 }
 
 static st_table *
-sym_hash(hdrp, syms)
-    struct exec *hdrp;
-    struct nlist *syms;
+sym_hash(struct exec *hdrp, struct nlist *syms)
 {
     st_table *tbl;
     struct nlist *sym = syms;
@@ -358,17 +345,16 @@ sym_hash(hdrp, syms)
 }
 
 static int
-dln_init(prog)
-    const char *prog;
+dln_init(const char *prog)
 {
-    char *file;
+    char *file, fbuf[MAXPATHLEN];
     int fd;
     struct exec hdr;
     struct nlist *syms;
 
     if (dln_init_p == 1) return 0;
 
-    file = dln_find_exe(prog, NULL);
+    file = dln_find_exe_r(prog, NULL, fbuf, sizeof(fbuf));
     if (file == NULL || (fd = open(file, O_RDONLY)) < 0) {
 	dln_errno = errno;
 	return -1;
@@ -433,11 +419,7 @@ dln_init(prog)
 }
 
 static long
-load_text_data(fd, hdrp, bss, disp)
-    int fd;
-    struct exec *hdrp;
-    int bss;
-    long disp;
+load_text_data(int fd, struct exec *hdrp, int bss, long disp)
 {
     int size;
     unsigned char* addr;
@@ -471,27 +453,26 @@ load_text_data(fd, hdrp, bss, disp)
 }
 
 static int
-undef_print(key, value)
-    char *key, *value;
+undef_print(char *key, char *value)
 {
     fprintf(stderr, "  %s\n", key);
     return ST_CONTINUE;
 }
 
 static void
-dln_print_undef()
+dln_print_undef(void)
 {
     fprintf(stderr, " Undefined symbols:\n");
     st_foreach(undef_tbl, undef_print, NULL);
 }
 
 static void
-dln_undefined()
+dln_undefined(void)
 {
     if (undef_tbl->num_entries > 0) {
 	fprintf(stderr, "dln: Calling undefined function\n");
 	dln_print_undef();
-	rb_exit(1);
+	dln_exit(1);
     }
 }
 
@@ -509,10 +490,7 @@ struct undef {
 
 static st_table *reloc_tbl = NULL;
 static void
-link_undef(name, base, reloc)
-    const char *name;
-    long base;
-    struct relocation_info *reloc;
+link_undef(const char *name, long base, struct relocation_info *reloc)
 {
     static int u_no = 0;
     struct undef *obj;
@@ -545,14 +523,11 @@ struct reloc_arg {
 };
 
 static int
-reloc_undef(no, undef, arg)
-    int no;
-    struct undef *undef;
-    struct reloc_arg *arg;
+reloc_undef(int no, struct undef *undef, struct reloc_arg *arg)
 {
     int datum;
     char *address;
-#if defined(sun) && defined(sparc)
+#if defined(__sun) && defined(__sparc)
     unsigned int mask = 0;
 #endif
 
@@ -561,7 +536,7 @@ reloc_undef(no, undef, arg)
     datum = arg->value;
 
     if (R_PCREL(&(undef->reloc))) datum -= undef->base;
-#if defined(sun) && defined(sparc)
+#if defined(__sun) && defined(__sparc)
     datum += undef->reloc.r_addend;
     datum >>= R_RIGHTSHIFT(&(undef->reloc));
     mask = (1 << R_BITSIZE(&(undef->reloc))) - 1;
@@ -609,9 +584,7 @@ reloc_undef(no, undef, arg)
 }
 
 static void
-unlink_undef(name, value)
-    const char *name;
-    long value;
+unlink_undef(const char *name, long value)
 {
     struct reloc_arg arg;
 
@@ -626,10 +599,7 @@ struct indr_data {
 };
 
 static int
-reloc_repl(no, undef, data)
-    int no;
-    struct undef *undef;
-    struct indr_data *data;
+reloc_repl(int no, struct undef *undef, struct indr_data *data)
 {
     if (strcmp(data->name0, undef->name) == 0) {
 	free(undef->name);
@@ -640,10 +610,7 @@ reloc_repl(no, undef, data)
 #endif
 
 static int
-load_1(fd, disp, need_init)
-    int fd;
-    long disp;
-    const char *need_init;
+load_1(int fd, long disp, const char *need_init)
 {
     static const char *libc = LIBC_NAME;
     struct exec hdr;
@@ -793,11 +760,11 @@ load_1(fd, disp, need_init)
 	while (rel < rel_end) {
 	    char *address = (char*)(rel->r_address + block);
 	    long datum = 0;
-#if defined(sun) && defined(sparc)
+#if defined(__sun) && defined(__sparc)
 	    unsigned int mask = 0;
 #endif
 
-	    if(rel >= rel_beg)
+	    if (rel >= rel_beg)
 		address += hdr.a_text;
 
 	    if (rel->r_extern) { /* Look it up in symbol-table */
@@ -814,7 +781,7 @@ load_1(fd, disp, need_init)
 		}
 	    } /* end.. look it up */
 	    else { /* is static */
-		switch (R_SYMBOL(rel)) { 
+		switch (R_SYMBOL(rel)) {
 		  case N_TEXT:
 		  case N_DATA:
 		    datum = block;
@@ -828,7 +795,7 @@ load_1(fd, disp, need_init)
 	    } /* end .. is static */
 	    if (R_PCREL(rel)) datum -= block;
 
-#if defined(sun) && defined(sparc)
+#if defined(__sun) && defined(__sparc)
 	    datum += rel->r_addend;
 	    datum >>= R_RIGHTSHIFT(rel);
 	    mask = (1 << R_BITSIZE(rel)) - 1;
@@ -924,10 +891,7 @@ load_1(fd, disp, need_init)
 
 static int target_offset;
 static int
-search_undef(key, value, lib_tbl)
-    const char *key;
-    int value;
-    st_table *lib_tbl;
+search_undef(const char *key, int value, st_table *lib_tbl)
 {
     long offset;
 
@@ -941,13 +905,13 @@ struct symdef {
     int lib_offset;
 };
 
-char *dln_librrb_ary_path = DLN_DEFAULT_LIB_PATH;
+const char *dln_librrb_ary_path = DLN_DEFAULT_LIB_PATH;
 
 static int
-load_lib(lib)
-    const char *lib;
+load_lib(const char *lib)
 {
-    char *path, *file;
+    char *path, *file, fbuf[MAXPATHLEN];
+    char *envpath = 0;
     char armagic[SARMAG];
     int fd, size;
     struct ar_hdr ahdr;
@@ -977,8 +941,10 @@ load_lib(lib)
     /* if path is still NULL, use "." for path. */
     path = getenv("DLN_LIBRARY_PATH");
     if (path == NULL) path = dln_librrb_ary_path;
+    else path = envpath = strdup(path);
 
-    file = dln_find_file(lib, path);
+    file = dln_find_file_r(lib, path, fbuf, sizeof(fbuf));
+    if (envpath) free(envpath);
     fd = open(file, O_RDONLY);
     if (fd == -1) goto syserr;
     size = read(fd, armagic, SARMAG);
@@ -1082,8 +1048,7 @@ load_lib(lib)
 }
 
 static int
-load(file)
-    const char *file;
+load(const char *file)
 {
     int fd;
     int result;
@@ -1108,8 +1073,7 @@ load(file)
 }
 
 void*
-dln_sym(name)
-    const char *name;
+dln_sym(const char *name)
 {
     struct nlist *sym;
 
@@ -1121,12 +1085,7 @@ dln_sym(name)
 #endif /* USE_DLN_A_OUT */
 
 #ifdef USE_DLN_DLOPEN
-# if defined(__NetBSD__) && defined(__NetBSD_Version__) && __NetBSD_Version__ < 105000000
-#  include <nlist.h>
-#  include <link.h>
-# else
-#  include <dlfcn.h>
-# endif
+# include <dlfcn.h>
 #endif
 
 #ifdef __hpux
@@ -1155,21 +1114,35 @@ dln_sym(name)
 #endif
 #endif
 
-#if defined _WIN32 && !defined __CYGWIN__
+#ifdef _WIN32
 #include <windows.h>
+#include <imagehlp.h>
 #endif
 
-#ifdef _WIN32_WCE
-#undef FormatMessage
-#define FormatMessage FormatMessageA
-#undef LoadLibrary
-#define LoadLibrary LoadLibraryA
-#undef GetProcAddress
-#define GetProcAddress GetProcAddressA
-#endif
-
+#ifdef _WIN32
 static const char *
-dln_strerror()
+dln_strerror(char *message, size_t size)
+{
+    int error = GetLastError();
+    char *p = message;
+    size_t len = snprintf(message, size, "%d: ", error);
+
+#define format_message(sublang) FormatMessage(\
+	FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,	\
+	NULL, error, MAKELANGID(LANG_NEUTRAL, (sublang)),		\
+	message + len, size - len, NULL)
+    if (format_message(SUBLANG_ENGLISH_US) == 0)
+	format_message(SUBLANG_DEFAULT);
+    for (p = message + len; *p; p++) {
+	if (*p == '\n' || *p == '\r')
+	    *p = ' ';
+    }
+    return message;
+}
+#define dln_strerror() dln_strerror(message, sizeof message)
+#elif ! defined _AIX
+static const char *
+dln_strerror(void)
 {
 #ifdef USE_DLN_A_OUT
     char *strerror();
@@ -1195,127 +1168,147 @@ dln_strerror()
 #ifdef USE_DLN_DLOPEN
     return (char*)dlerror();
 #endif
-
-#if defined _WIN32 && !defined __CYGWIN__
-    static char message[1024];
-    int error = GetLastError();
-    char *p = message;
-    p += sprintf(message, "%d: ", error);
-    FormatMessage(
-	FORMAT_MESSAGE_FROM_SYSTEM	 | FORMAT_MESSAGE_IGNORE_INSERTS,
-	NULL,
-	error,
-	MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
-	p,
-	sizeof message - strlen(message),
-	NULL);
-
-    for (p = message; *p; p++) {
-	if (*p == '\n' || *p == '\r')
-	    *p = ' ';
-    }
-    return message;
-#endif
 }
-
+#endif
 
 #if defined(_AIX) && ! defined(_IA64)
 static void
 aix_loaderror(const char *pathname)
 {
-    char *message[8], errbuf[1024];
-    int i,j;
+    char *message[1024], errbuf[1024];
+    int i;
+#define ERRBUF_APPEND(s) strlcat(errbuf, (s), sizeof(errbuf))
+    snprintf(errbuf, sizeof(errbuf), "load failed - %s. ", pathname);
 
-    struct errtab { 
-	int errnum;
-	char *errstr;
-    } load_errtab[] = {
-	{L_ERROR_TOOMANY,	"too many errors, rest skipped."},
-	{L_ERROR_NOLIB,		"can't load library:"},
-	{L_ERROR_UNDEF,		"can't find symbol in library:"},
-	{L_ERROR_RLDBAD,
-	     "RLD index out of range or bad relocation type:"},
-	{L_ERROR_FORMAT,	"not a valid, executable xcoff file:"},
-	{L_ERROR_MEMBER,
-	     "file not an archive or does not contain requested member:"},
-	{L_ERROR_TYPE,		"symbol table mismatch:"},
-	{L_ERROR_ALIGN,		"text alignment in file is wrong."},
-	{L_ERROR_SYSTEM,	"System error:"},
-	{L_ERROR_ERRNO,		NULL}
-    };
-
-#define LOAD_ERRTAB_LEN	(sizeof(load_errtab)/sizeof(load_errtab[0]))
-#define ERRBUF_APPEND(s) strncat(errbuf, s, sizeof(errbuf)-strlen(errbuf)-1)
-
-    snprintf(errbuf, 1024, "load failed - %s ", pathname);
-
-    if (!loadquery(1, &message[0], sizeof(message))) 
-	ERRBUF_APPEND(strerror(errno));
-    for(i = 0; message[i] && *message[i]; i++) {
-	int nerr = atoi(message[i]);
-	for (j=0; j<LOAD_ERRTAB_LEN; j++) {
-           if (nerr == load_errtab[i].errnum && load_errtab[i].errstr)
-		ERRBUF_APPEND(load_errtab[i].errstr);
+    if (loadquery(L_GETMESSAGES, &message[0], sizeof(message)) != -1) {
+	ERRBUF_APPEND("Please issue below command for detailed reasons:\n\t");
+	ERRBUF_APPEND("/usr/sbin/execerror ruby ");
+	for (i=0; message[i]; i++) {
+	    ERRBUF_APPEND("\"");
+	    ERRBUF_APPEND(message[i]);
+	    ERRBUF_APPEND("\" ");
 	}
-	while (isdigit(*message[i])) message[i]++; 
-	ERRBUF_APPEND(message[i]);
 	ERRBUF_APPEND("\n");
     }
-    errbuf[strlen(errbuf)-1] = '\0';	/* trim off last newline */
-    rb_loaderror(errbuf);
-    return;
+    else {
+	ERRBUF_APPEND(strerror(errno));
+	ERRBUF_APPEND("[loadquery failed]");
+    }
+    dln_loaderror("%s", errbuf);
 }
 #endif
 
-#if defined(__VMS)
-#include <starlet.h>
-#include <rms.h>
-#include <stsdef.h>
-#include <unixlib.h>
-#include <descrip.h>
-#include <lib$routines.h>
+#if defined _WIN32 && defined RUBY_EXPORT
+HANDLE rb_libruby_handle(void);
 
-static char *vms_filespec;
-static int vms_fileact(char *filespec, int type);
-static long vms_fisexh(long *sigarr, long *mecarr);
+static int
+rb_w32_check_imported(HMODULE ext, HMODULE mine)
+{
+    ULONG size;
+    const IMAGE_IMPORT_DESCRIPTOR *desc;
+
+    desc = ImageDirectoryEntryToData(ext, TRUE, IMAGE_DIRECTORY_ENTRY_IMPORT, &size);
+    if (!desc) return 0;
+    while (desc->Name) {
+	PIMAGE_THUNK_DATA pint = (PIMAGE_THUNK_DATA)((char *)ext + desc->Characteristics);
+	PIMAGE_THUNK_DATA piat = (PIMAGE_THUNK_DATA)((char *)ext + desc->FirstThunk);
+	for (; piat->u1.Function; piat++, pint++) {
+	    static const char prefix[] = "rb_";
+	    PIMAGE_IMPORT_BY_NAME pii;
+	    const char *name;
+
+	    if (IMAGE_SNAP_BY_ORDINAL(pint->u1.Ordinal)) continue;
+	    pii = (PIMAGE_IMPORT_BY_NAME)((char *)ext + (size_t)pint->u1.AddressOfData);
+	    name = (const char *)pii->Name;
+	    if (strncmp(name, prefix, sizeof(prefix) - 1) == 0) {
+		FARPROC addr = GetProcAddress(mine, name);
+		if (addr) return (FARPROC)piat->u1.Function == addr;
+	    }
+	}
+	desc++;
+    }
+    return 1;
+}
 #endif
 
-#endif /* NO_DLN_LOAD */
+#if defined(DLN_NEEDS_ALT_SEPARATOR) && DLN_NEEDS_ALT_SEPARATOR
+#define translit_separator(src) do { \
+	char *tmp = ALLOCA_N(char, strlen(src) + 1), *p = tmp, c; \
+	do { \
+	    *p++ = ((c = *file++) == '/') ? DLN_NEEDS_ALT_SEPARATOR : c; \
+	} while (c); \
+	(src) = tmp; \
+    } while (0)
+#else
+#define translit_separator(str) (void)(str)
+#endif
+
+#ifdef USE_DLN_DLOPEN
+#ifdef __clang__
+#pragma clang diagnostic push
+#pragma clang diagnostic ignored "-Wpedantic"
+#elif defined(__GNUC__) && (__GNUC__ >= 5)
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wpedantic"
+#endif
+static bool
+dln_incompatible_library_p(void *handle)
+{
+    void *ex = dlsym(handle, EXTERNAL_PREFIX"ruby_xmalloc");
+    return ex && ex != ruby_xmalloc;
+}
+#ifdef __clang__
+#pragma clang diagnostic pop
+#elif defined(__GNUC__) && (__GNUC__ >= 5)
+#pragma GCC diagnostic pop
+#endif
+#endif
 
 void*
-dln_load(file)
-    const char *file;
+dln_load(const char *file)
 {
-#ifdef NO_DLN_LOAD
-    rb_raise(rb_eLoadError, "this executable file can't load extension libraries");
-#else
-
+#if (defined _WIN32 || defined USE_DLN_DLOPEN) && defined RUBY_EXPORT
+    static const char incompatible[] = "incompatible library version";
+#endif
 #if !defined(_AIX) && !defined(NeXT)
     const char *error = 0;
-#define DLN_ERROR() (error = dln_strerror(), strcpy(ALLOCA_N(char, strlen(error) + 1), error))
 #endif
 
-#if defined _WIN32 && !defined __CYGWIN__
+#if defined _WIN32
     HINSTANCE handle;
-    char winfile[MAXPATHLEN];
+    WCHAR *winfile;
+    char message[1024];
     void (*init_fct)();
     char *buf;
 
-    if (strlen(file) >= MAXPATHLEN) rb_loaderror("filename too long");
-
     /* Load the file as an object one */
     init_funcname(&buf, file);
 
-    strcpy(winfile, file);
+    /* Convert the file path to wide char */
+    winfile = rb_w32_mbstr_to_wstr(CP_UTF8, file, -1, NULL);
+    if (!winfile) {
+	dln_memerror();
+    }
 
     /* Load file */
-    if ((handle = LoadLibrary(winfile)) == NULL) {
+    handle = LoadLibraryW(winfile);
+    free(winfile);
+
+    if (!handle) {
 	error = dln_strerror();
 	goto failed;
     }
 
+#if defined _WIN32 && defined RUBY_EXPORT
+    if (!rb_w32_check_imported(handle, rb_libruby_handle())) {
+	FreeLibrary(handle);
+	error = incompatible;
+	goto failed;
+    }
+#endif
+
     if ((init_fct = (void(*)())GetProcAddress(handle, buf)) == NULL) {
-	rb_loaderror("%s - %s\n%s", dln_strerror(), buf, file);
+	dln_loaderror("%s - %s\n%s", dln_strerror(), buf, file);
     }
 
     /* Call the init code */
@@ -1333,6 +1326,7 @@ dln_load(file)
     char *buf;
     /* Load the file as an object one */
     init_funcname(&buf, file);
+    translit_separator(file);
 
 #ifdef USE_DLN_DLOPEN
 #define DLN_DEFINED
@@ -1355,10 +1349,28 @@ dln_load(file)
 	    error = dln_strerror();
 	    goto failed;
 	}
+# if defined RUBY_EXPORT
+	{
+	    if (dln_incompatible_library_p(handle)) {
 
-	init_fct = (void(*)())dlsym(handle, buf);
+#   if defined __APPLE__ && \
+    defined(MAC_OS_X_VERSION_MIN_REQUIRED) && \
+    (MAC_OS_X_VERSION_MIN_REQUIRED < MAC_OS_X_VERSION_10_11)
+		/* dlclose() segfaults */
+		rb_fatal("%s - %s", incompatible, file);
+#   else
+		dlclose(handle);
+		error = incompatible;
+		goto failed;
+#   endif
+	    }
+	}
+# endif
+
+	init_fct = (void(*)())(VALUE)dlsym(handle, buf);
 	if (init_fct == NULL) {
-	    error = DLN_ERROR();
+	    const size_t errlen = strlen(error = dln_strerror()) + 1;
+	    error = memcpy(ALLOCA_N(char, errlen), error, errlen);
 	    dlclose(handle);
 	    goto failed;
 	}
@@ -1380,14 +1392,14 @@ dln_load(file)
 	lib = shl_load(file, flags, 0);
 	if (lib == NULL) {
 	    extern int errno;
-	    rb_loaderror("%s - %s", strerror(errno), file);
+	    dln_loaderror("%s - %s", strerror(errno), file);
 	}
 	shl_findsym(&lib, buf, TYPE_PROCEDURE, (void*)&init_fct);
 	if (init_fct == NULL) {
 	    shl_findsym(&lib, buf, TYPE_UNDEFINED, (void*)&init_fct);
 	    if (init_fct == NULL) {
 		errno = ENOSYM;
-		rb_loaderror("%s - %s", strerror(ENOSYM), file);
+		dln_loaderror("%s - %s", strerror(ENOSYM), file);
 	    }
 	}
 	(*init_fct)();
@@ -1412,54 +1424,17 @@ dln_load(file)
     }
 #endif /* _AIX */
 
-#if defined(NeXT) || defined(MACOSX_DYLD)
+#if defined(MACOSX_DYLD)
 #define DLN_DEFINED
 /*----------------------------------------------------
    By SHIROYAMA Takayuki Psi@fortune.nest.or.jp
- 
+
    Special Thanks...
     Yu tomoak-i@is.aist-nara.ac.jp,
     Mi hisho@tasihara.nest.or.jp,
     sunshine@sunshineco.com,
     and... Miss ARAI Akino(^^;)
  ----------------------------------------------------*/
-#if defined(NeXT) && (NS_TARGET_MAJOR < 4)/* NeXTSTEP rld functions */
-
-    {
-        NXStream* s;
-	unsigned long init_address;
-	char *object_files[2] = {NULL, NULL};
-
-	void (*init_fct)();
-	
-	object_files[0] = (char*)file;
-	
-	s = NXOpenFile(2,NX_WRITEONLY);
-
-	/* Load object file, if return value ==0 ,  load failed*/
-	if(rld_load(s, NULL, object_files, NULL) == 0) {
-	    NXFlush(s);
-	    NXClose(s);
-	    rb_loaderror("Failed to load %.200s", file);
-	}
-
-	/* lookup the initial function */
-	if(rld_lookup(s, buf, &init_address) == 0) {
-	    NXFlush(s);
-	    NXClose(s);
-	    rb_loaderror("Failed to lookup Init function %.200s", file);
-	}
-
-	NXFlush(s);
-	NXClose(s);
-
-	/* Cannot call *init_address directory, so copy this value to
-	   funtion pointer */
-	init_fct = (void(*)())init_address;
-	(*init_fct)();
-	return (void*)init_address;
-    }
-#else/* OPENSTEP dyld functions */
     {
 	int dyld_result;
 	NSObjectFileImage obj_file; /* handle, but not use it */
@@ -1472,398 +1447,32 @@ dln_load(file)
 	dyld_result = NSCreateObjectFileImageFromFile(file, &obj_file);
 
 	if (dyld_result != NSObjectFileImageSuccess) {
-	    rb_loaderror("Failed to load %.200s", file);
+	    dln_loaderror("Failed to load %.200s", file);
 	}
 
 	NSLinkModule(obj_file, file, NSLINKMODULE_OPTION_BINDNOW);
 
 	/* lookup the initial function */
-	if(!NSIsSymbolNameDefined(buf)) {
-	    rb_loaderror("Failed to lookup Init function %.200s",file);
-	}	
+	if (!NSIsSymbolNameDefined(buf)) {
+	    dln_loaderror("Failed to lookup Init function %.200s",file);
+	}
 	init_fct = NSAddressOfSymbol(NSLookupAndBindSymbol(buf));
 	(*init_fct)();
 
 	return (void*)init_fct;
     }
-#endif /* rld or dyld */
 #endif
 
-#ifdef __BEOS__
-# define DLN_DEFINED
-    {
-      status_t err_stat;  /* BeOS error status code */
-      image_id img_id;    /* extention module unique id */
-      void (*init_fct)(); /* initialize function for extention module */
-
-      /* load extention module */
-      img_id = load_add_on(file);
-      if (img_id <= 0) {
-	rb_loaderror("Failed to load %.200s", file);
-      }
-      
-      /* find symbol for module initialize function. */
-      /* The Be Book KernelKit Images section described to use
-	 B_SYMBOL_TYPE_TEXT for symbol of function, not
-	 B_SYMBOL_TYPE_CODE. Why ? */
-      /* strcat(init_fct_symname, "__Fv"); */  /* parameter nothing. */
-      /* "__Fv" dont need! The Be Book Bug ? */
-      err_stat = get_image_symbol(img_id, buf,
-				  B_SYMBOL_TYPE_TEXT, (void **)&init_fct);
-
-      if (err_stat != B_NO_ERROR) {
-	char real_name[MAXPATHLEN];
-
-	strcpy(real_name, buf);
-	strcat(real_name, "__Fv");
-        err_stat = get_image_symbol(img_id, real_name,
-				    B_SYMBOL_TYPE_TEXT, (void **)&init_fct);
-      }
-
-      if ((B_BAD_IMAGE_ID == err_stat) || (B_BAD_INDEX == err_stat)) {
-	unload_add_on(img_id);
-	rb_loaderror("Failed to lookup Init function %.200s", file);
-      }
-      else if (B_NO_ERROR != err_stat) {
-	char errmsg[] = "Internal of BeOS version. %.200s (symbol_name = %s)";
-	unload_add_on(img_id);
-	rb_loaderror(errmsg, strerror(err_stat), buf);
-      }
-
-      /* call module initialize function. */
-      (*init_fct)();
-      return (void*)img_id;
-    }
-#endif /* __BEOS__*/
-
-#ifdef __MACOS__   /* Mac OS 9 or before */
-# define DLN_DEFINED
-    {
-      OSErr err;
-      FSSpec libspec;
-      CFragConnectionID connID;
-      Ptr mainAddr;
-      char errMessage[1024];
-      Boolean isfolder, didsomething;
-      Str63 fragname;
-      Ptr symAddr;
-      CFragSymbolClass class;
-      void (*init_fct)();
-      char fullpath[MAXPATHLEN];
-
-      strcpy(fullpath, file);
-
-      /* resolve any aliases to find the real file */
-      c2pstr(fullpath);
-      (void)FSMakeFSSpec(0, 0, fullpath, &libspec);
-      err = ResolveAliasFile(&libspec, 1, &isfolder, &didsomething);
-      if (err) {
-	  rb_loaderror("Unresolved Alias - %s", file);
-      }
-
-      /* Load the fragment (or return the connID if it is already loaded */
-      fragname[0] = 0;
-      err = GetDiskFragment(&libspec, 0, 0, fragname, 
-			    kLoadCFrag, &connID, &mainAddr,
-			    errMessage);
-      if (err) {
-	  p2cstr(errMessage);
-	  rb_loaderror("%s - %s",errMessage , file);
-      }
-
-      /* Locate the address of the correct init function */
-      c2pstr(buf);
-      err = FindSymbol(connID, buf, &symAddr, &class);
-      if (err) {
-	  rb_loaderror("Unresolved symbols - %s" , file);
-      }
-      init_fct = (void (*)())symAddr;
-      (*init_fct)();
-      return (void*)init_fct;
-    }
-#endif /* __MACOS__ */
-
-#if defined(__VMS)
-#define DLN_DEFINED
-    {
-	long status;
-	void (*init_fct)();
-	char *fname, *p1, *p2;
-
-	$DESCRIPTOR(fname_d, "");
-	$DESCRIPTOR(image_d, "");
-	$DESCRIPTOR(buf_d, "");
-
-	decc$to_vms(file, vms_fileact, 0, 0);
-
-	fname = (char *)__alloca(strlen(file)+1);
-	strcpy(fname,file);
-	if (p1 = strrchr(fname,'/'))
-	    fname = p1 + 1;
-	if (p2 = strrchr(fname,'.'))
-	    *p2 = '\0';
-
-	fname_d.dsc$w_length  = strlen(fname);
-	fname_d.dsc$a_pointer = fname;
-	image_d.dsc$w_length  = strlen(vms_filespec);
-	image_d.dsc$a_pointer = vms_filespec;
-	buf_d.dsc$w_length    = strlen(buf);
-	buf_d.dsc$a_pointer   = buf;
-
-	lib$establish(vms_fisexh);
-
-	status = lib$find_image_symbol (
-		     &fname_d,
-		     &buf_d, 
-		     &init_fct, 
-		     &image_d);
-
-	lib$establish(0);
-
-	if (status == RMS$_FNF) {
-	    error = dln_strerror();
-	    goto failed;
-	} else if (!$VMS_STATUS_SUCCESS(status)) {
-	    error = DLN_ERROR();
-	    goto failed;
-	}
-
-	/* Call the init code */
-	(*init_fct)();
-
-	return 1;
-    }
-#endif /* __VMS */
-
 #ifndef DLN_DEFINED
-    rb_notimplement();
+    dln_notimplement();
 #endif
 
 #endif /* USE_DLN_A_OUT */
 #endif
 #if !defined(_AIX) && !defined(NeXT)
   failed:
-    rb_loaderror("%s - %s", error, file);
+    dln_loaderror("%s - %s", error, file);
 #endif
 
-#endif /* NO_DLN_LOAD */
     return 0;			/* dummy return */
 }
-
-static char *dln_find_1();
-
-char *
-dln_find_exe(fname, path)
-    const char *fname;
-    const char *path;
-{
-    if (!path) {
-	path = getenv(PATH_ENV);
-    }
-
-    if (!path) {
-#if defined(MSDOS) || defined(_WIN32) || defined(__human68k__) || defined(__MACOS__)
-	path = "/usr/local/bin;/usr/ucb;/usr/bin;/bin;.";
-#else
-	path = "/usr/local/bin:/usr/ucb:/usr/bin:/bin:.";
-#endif
-    }
-    return dln_find_1(fname, path, 1);
-}
-
-char *
-dln_find_file(fname, path)
-    const char *fname;
-    const char *path;
-{
-#ifndef __MACOS__
-    if (!path) path = ".";
-    return dln_find_1(fname, path, 0);
-#else
-    if (!path) path = ".";
-    return _macruby_path_conv_posix_to_macos(dln_find_1(fname, path, 0));
-#endif
-}
-
-static char fbuf[MAXPATHLEN];
-
-static char *
-dln_find_1(fname, path, exe_flag)
-    char *fname;
-    char *path;
-    int exe_flag;		/* non 0 if looking for executable. */
-{
-    register char *dp;
-    register char *ep;
-    register char *bp;
-    struct stat st;
-#ifdef __MACOS__
-    const char* mac_fullpath;
-#endif
-
-    if (!fname) return fname;
-    if (fname[0] == '/') return fname;
-    if (strncmp("./", fname, 2) == 0 || strncmp("../", fname, 3) == 0)
-      return fname;
-    if (exe_flag && strchr(fname, '/')) return fname;
-#ifdef DOSISH
-    if (fname[0] == '\\') return fname;
-# ifdef DOSISH_DRIVE_LETTER
-    if (strlen(fname) > 2 && fname[1] == ':') return fname;
-# endif
-    if (strncmp(".\\", fname, 2) == 0 || strncmp("..\\", fname, 3) == 0)
-      return fname;
-    if (exe_flag && strchr(fname, '\\')) return fname;
-#endif
-
-    for (dp = path;; dp = ++ep) {
-	register int l;
-	int i;
-	int fspace;
-
-	/* extract a component */
-	ep = strchr(dp, PATH_SEP[0]);
-	if (ep == NULL)
-	    ep = dp+strlen(dp);
-
-	/* find the length of that component */
-	l = ep - dp;
-	bp = fbuf;
-	fspace = sizeof fbuf - 2;
-	if (l > 0) {
-	    /*
-	    **	If the length of the component is zero length,
-	    **	start from the current directory.  If the
-	    **	component begins with "~", start from the
-	    **	user's $HOME environment variable.  Otherwise
-	    **	take the path literally.
-	    */
-
-	    if (*dp == '~' && (l == 1 ||
-#if defined(DOSISH)
-			       dp[1] == '\\' || 
-#endif
-			       dp[1] == '/')) {
-		char *home;
-
-		home = getenv("HOME");
-		if (home != NULL) {
-		    i = strlen(home);
-		    if ((fspace -= i) < 0)
-			goto toolong;
-		    memcpy(bp, home, i);
-		    bp += i;
-		}
-		dp++;
-		l--;
-	    }
-	    if (l > 0) {
-		if ((fspace -= l) < 0)
-		    goto toolong;
-		memcpy(bp, dp, l);
-		bp += l;
-	    }
-
-	    /* add a "/" between directory and filename */
-	    if (ep[-1] != '/')
-		*bp++ = '/';
-	}
-
-	/* now append the file name */
-	i = strlen(fname);
-	if ((fspace -= i) < 0) {
-	  toolong:
-	    fprintf(stderr, "openpath: pathname too long (ignored)\n");
-	    *bp = '\0';
-	    fprintf(stderr, "\tDirectory \"%s\"\n", fbuf);
-	    fprintf(stderr, "\tFile \"%s\"\n", fname);
-	    goto next;
-	}
-	memcpy(bp, fname, i + 1);
-
-#ifndef __MACOS__
-	if (stat(fbuf, &st) == 0) {
-	    if (exe_flag == 0) return fbuf;
-	    /* looking for executable */
-	    if (!S_ISDIR(st.st_mode) && eaccess(fbuf, X_OK) == 0)
-		return fbuf;
-	}
-#else
-	if (mac_fullpath = _macruby_exist_file_in_libdir_as_posix_name(fbuf)) {
-	    if (exe_flag == 0) return mac_fullpath;
-	    /* looking for executable */
-	    if (stat(mac_fullpath, &st) == 0) {
-		if (!S_ISDIR(st.st_mode) && eaccess(mac_fullpath, X_OK) == 0)
-		    return mac_fullpath;
-	    }
-	}
-#endif
-#if defined(DOSISH)
-	if (exe_flag) {
-	    static const char *const extension[] = {
-#if defined(MSDOS)
-		".com", ".exe", ".bat",
-#if defined(DJGPP)
-		".btm", ".sh", ".ksh", ".pl", ".sed",
-#endif
-#elif defined(__EMX__) || defined(_WIN32)
-		".exe", ".com", ".cmd", ".bat",
-/* end of __EMX__ or _WIN32 */
-#else
-		".r", ".R", ".x", ".X", ".bat", ".BAT",
-/* __human68k__ */
-#endif
-		(char *) NULL
-	    };
-	    int j;
-
-	    for (j = 0; extension[j]; j++) {
-		if (fspace < strlen(extension[j])) {
-		    fprintf(stderr, "openpath: pathname too long (ignored)\n");
-		    fprintf(stderr, "\tDirectory \"%.*s\"\n", (int) (bp - fbuf), fbuf);
-		    fprintf(stderr, "\tFile \"%s%s\"\n", fname, extension[j]);
-		    continue;
-		}
-		strcpy(bp + i, extension[j]);
-#ifndef __MACOS__
-		if (stat(fbuf, &st) == 0)
-		    return fbuf;
-#else
-		if (mac_fullpath = _macruby_exist_file_in_libdir_as_posix_name(fbuf))
-		    return mac_fullpath;
-
-#endif
-	    }
-	}
-#endif /* MSDOS or _WIN32 or __human68k__ or __EMX__ */
-
-      next:
-	/* if not, and no other alternatives, life is bleak */
-	if (*ep == '\0') {
-	    return NULL;
-	}
-
-	/* otherwise try the next component in the search path */
-    }
-}
-
-#if defined(__VMS)
-
-/* action routine for decc$to_vms */
-static int vms_fileact(char *filespec, int type)
-{
-    if (vms_filespec)
-	free(vms_filespec);
-    vms_filespec = malloc(strlen(filespec)+1);
-    strcpy(vms_filespec, filespec);
-    return 1;
-}
-
-/* exception handler for LIB$FIND_IMAGE_SYMBOL */
-static long vms_fisexh(long *sigarr, long *mecarr)
-{
-    sys$unwind(1, 0);
-    return 1;
-}
-
-#endif /* __VMS */
